"""
Template configuration tools for Spira by Inflectra
"""
from mcp_server_spira.features.templateconfiguration.tools import (
    artifacttypes, customproperties
)


def register_tools(mcp) -> None:
    """
    Register all template configurationspace tools with the MCP server.
    
    Args:
        mcp: The FastMCP server instance
    """
    artifacttypes.register_tools(mcp)
    customproperties.register_tools(mcp)
    


"""
Provides operations for retrieving the product specification files that
can be used to build the functionality of the product using AI. 
This is used by Agentic AI development tools such as Amazon Kiro
for building applications from a formal spec.

A product specification consists of the data in markdown format used by Kiro to generate
the following files:
    - requirements.md - Captures user stories and acceptance criteria in structured EARS notation
    - design.md - Documents technical architecture, sequence diagrams, and implementation considerations
    - tasks.md - Provides a detailed implementation plan with discrete, trackable tasks
    - test-cases.md - Includes detailed test scenarios and steps, expected results and pass/fail criteria

This module provides the following MCP tools for retrieving the entire product specifications:
    - get_specification_requirements - returns the data for populating the requirements.md file
    - get_specification_design - returns the data for populating the design.md file
    - get_specification_tasks - returns the data for populating the tasks.md file
    - get_specification_test_cases - returns the data for populating the test-cases.md file
"""

from mcp.server.fastmcp.utilities.logging import get_logger
from mcp_server_spira.features.common import get_spira_client
from typing import Any

# Get a logger instance, typically named after the current module
logger = get_logger(__name__)

def _get_product_by_id(spira_client, product_id: int) -> Any:
    """
    Implementation of retrieving a single Spira product by its ID

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45.
                
    Returns:
        The product object from Spira
    """
    try:
        # Get the product by its ID
        product_url = f"projects/{product_id}"
        product = spira_client.make_spira_api_get_request(product_url)

        if not product:
            return "There was no product with that ID available"

        return product
    except Exception as e:
        raise e
    
def _get_release_by_id(spira_client, product_id: int, release_id: int) -> Any:
    """
    Retrieves a single release in the specified product with the specified ID

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
        release_id: The numeric ID of the release. If the ID is RL:12, just use 12.
                
    Returns:
        The Spira release object
    """
    try:
        # Get the release in the product
        release_url = f"projects/{product_id}/releases/{release_id}"
        release = spira_client.make_spira_api_get_request(release_url)

        if not release:
            return "There is no release with the specified ID."
        
        # Return the object
        return release        
    except Exception as e:
        raise e

def _get_specification_requirements(spira_client, product_id: int, release_id: int | None) -> list[Any]:
    """
    Gets the list of requirements in the product/release

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
        release_id: The numeric ID of the release. If the ID is RL:12, just use 12. (optional)
                
    Returns:
        List of requirements
    """
    try:
        requirements = []
        starting_row = 1
        number_of_rows = 250
        more_results = True

        # See if we are filtering by release or not
        if release_id:
            while more_results:
                requirements_url = f"projects/{product_id}/requirements/search?starting_row={starting_row}&number_of_rows={number_of_rows}"
                body = [{'PropertyName': 'ReleaseId', 'IntValue': release_id}]
                results = spira_client.make_spira_api_post_request(requirements_url, body)
                if not results:
                    more_results = False
                else:
                    starting_row += number_of_rows
                requirements.extend(results)
        else:
            while more_results:
                requirements_url = f"projects/{product_id}/requirements?starting_row={starting_row}&number_of_rows={number_of_rows}"
                results = spira_client.make_spira_api_get_request(requirements_url)
                if not results:
                    more_results = False
                else:
                    starting_row += number_of_rows
                requirements.extend(results)

        return requirements
    except Exception as e:
        raise e

def _add_requirement_scenarios(spira_client, product_id: int, requirement_id: int, formatted_specification: list[str]):
    """
    Gets the list of scenarios for a requirement and adds them to the output

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
        requirement_id: The numeric ID of the requirement. If the ID is RQ:12, just use 12
        formatted_specification: The output text in markdown format
    """
    scenarios_url = f"projects/{product_id}/requirements/{requirement_id}/steps"
    scenarios = spira_client.make_spira_api_get_request(scenarios_url)
    if scenarios:
        formatted_specification.append('#### Acceptance Criteria\n\n')
        for scenario in scenarios:
            position = scenario['Position']
            description =scenario['Description']
            text = f"{position}. {description}\n"
            formatted_specification.append(text)
        formatted_specification.append('\n')

def _add_requirement_test_cases(spira_client, product_id: int, requirement_id: int, formatted_specification: list[str]):
    """
    Gets the list of test cases for a requirement and adds them to the output

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
        requirement_id: The numeric ID of the requirement. If the ID is RQ:12, just use 12
        formatted_specification: The output text in markdown format
    """
    req_test_cases_url = f"projects/{product_id}/requirements/{requirement_id}/test-cases"
    req_test_cases = spira_client.make_spira_api_get_request(req_test_cases_url)
    if req_test_cases:
        formatted_specification.append('#### Test Cases\n\n')
        for req_test_case in req_test_cases:
            test_case_id = req_test_case['TestCaseId']

            # Get the full details of the test case
            test_case_url = f"projects/{product_id}/test-cases/{test_case_id}"
            test_case = spira_client.make_spira_api_get_request(test_case_url)
            
            if test_case:
                name = test_case['Name']
                formatted_specification.append(f"##### Test Case TC:{test_case_id}: {test_case['Name']}\n")
                if test_case['Description']:
                    description = f"**{test_case['TestCaseTypeName']}:** {test_case['Description']}\n"
                    formatted_specification.append(description)    
                formatted_specification.append('\n')

                # Get the test case steps
                test_steps_url = f"projects/{product_id}/test-cases/{test_case_id}/test-steps"
                test_steps = spira_client.make_spira_api_get_request(test_steps_url)

                # Format the test steps as a table
                if test_steps:
                    formatted_specification.append('##### Steps\n\n')
                    formatted_specification.append("<table>")
                    formatted_specification.append("<tr><th>Step #</th><th>Description</th><th>Expected Result</th><th>Sample Data</th></tr>")
                    for test_step in test_steps:
                        test_step_id = test_step['TestStepId']
                        position = test_step['Position']
                        description = test_step['Description']
                        expected_result = test_step['ExpectedResult']
                        sample_data = test_step['SampleData']
                        formatted_specification.append("<tr>")
                        formatted_specification.append(f"<td>{position}.</td>")
                        formatted_specification.append(f"<td>{description}.</td>")
                        formatted_specification.append(f"<td>{expected_result}.</td>")
                        formatted_specification.append(f"<td>{sample_data}.</td>")
                        formatted_specification.append("</tr>")
                    formatted_specification.append("</table>")

        formatted_specification.append('\n')

def _add_requirement_tasks(spira_client, product_id: int, requirement_id: int, formatted_specification: list[str]):
    """
    Gets the list of tasks for a requirement and adds them to the output

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
        requirement_id: The numeric ID of the requirement. If the ID is RQ:12, just use 12
        formatted_specification: The output text in markdown format
    """

    # First, get the count of the number of total matching tasks
    body = [{'PropertyName': 'RequirementId', 'IntValue': requirement_id}]
    tasks_url = f"projects/{product_id}/tasks/count "
    task_count = spira_client.make_spira_api_post_request(tasks_url, body)

    # Next, get all of the tasks using pagination
    tasks = []
    starting_row = 1
    number_of_rows = 250
    while starting_row < task_count:
        tasks_url = f"projects/{product_id}/tasks?starting_row={starting_row}&number_of_rows={number_of_rows}&sort_field=StartDate&sort_direction=ASC"
        results = spira_client.make_spira_api_post_request(tasks_url, body)
        starting_row += number_of_rows
        tasks.extend(results)

    if tasks:
        formatted_specification.append('#### Tasks\n\n')
        for task in tasks:
            task_id = task['TaskId']            
            name = task['Name']
            formatted_specification.append(f"##### Task TK:{task_id}: {task['Name']}\n")
            if task['Description']:
                description = f"**{task['TaskTypeName']}:** {task['Description']}\n"
                formatted_specification.append(description)    
            formatted_specification.append('\n')

        formatted_specification.append('\n')

def _get_specification_risks(spira_client, product_id: int, release_id: int | None) -> list[Any]:
    """
    Gets the list of risks in the product/release

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
        release_id: The numeric ID of the release. If the ID is RL:12, just use 12. (optional)
                
    Returns:
        List of risks
    """
    try:
        risks = []
        starting_row = 1
        number_of_rows = 250
        more_results = True
        body = []

        # See if we are filtering by release or not
        if release_id:
            body = [{'PropertyName': 'ReleaseId', 'IntValue': release_id}]

        while more_results:
            risks_url = f"projects/{product_id}/risks?starting_row={starting_row}&number_of_rows={number_of_rows}&sort_field=RiskExposure&sort_direction=DESC"
            results = spira_client.make_spira_api_post_request(risks_url, body)
            if not results:
                more_results = False
            else:
                starting_row += number_of_rows
            risks.extend(results)


        return risks
    except Exception as e:
        raise e

def _add_risk_mitigations(spira_client, product_id: int, risk_id: int, formatted_specification: list[str]):
    """
    Gets the list of mitigations for a risk and adds them to the output

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
        risk_id: The numeric ID of the risk. If the ID is RK:12, just use 12
        formatted_specification: The output text in markdown format
    """
    mitigations_url = f"projects/{product_id}/risks/{risk_id}/mitigations"
    mitigations = spira_client.make_spira_api_get_request(mitigations_url)
    if mitigations:
        formatted_specification.append('#### Mitigations\n\n')
        for mitigation in mitigations:
            position = mitigation['Position']
            description =mitigation['Description']
            text = f"{position}. {description}\n"
            formatted_specification.append(text)
        formatted_specification.append('\n')

def _get_specification_requirements_impl(spira_client, product_id: int, release_id: int | None) -> str:
    """
    Implementation of retrieving the requirements markdown specification for the specified product

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
        release_id: The numeric ID of the release. If the ID is RL:12, just use 12.
                    If no release is specified, then the specification for the entire
                    project is returned 
                
    Returns:
        Formatted string containing the product requirements specification
    """
    try:
        formatted_specification = []

        # Get the product information
        logger.info("Getting the product overview")
        product = _get_product_by_id(spira_client, product_id)
        product_name = product['Name']

        # Create the header
        if release_id:
            # Get the release information
            release = _get_release_by_id(spira_client, product_id, release_id)
            release_version_number = release['VersionNumber']
            formatted_specification.append(f'# Specification for {product_name} [PR:{product_id}], Release {release_version_number} [RL:{release_id}]')
        else:
            formatted_specification.append(f'# Specification for {product_name} [PR:{product_id}]')
        formatted_specification.append('\n')

        # Populate the product overview
        if product['Description']:
            formatted_specification.append(f'## Product Overview')
            formatted_specification.append(product['Description'])
            formatted_specification.append('\n')

        # Create the sub-header for the Requirements.md section
        formatted_specification.append('\n')
        formatted_specification.append(f'## Requirements Document')

        # Get the list of requirements in the product, or just the release
        requirements = _get_specification_requirements(spira_client, product_id, release_id)

        if requirements:
            # Format the requirements into human readable data
            for requirement in requirements:
                requirement_id = requirement['RequirementId']
                formatted_specification.append(f"### Requirement RQ:{requirement_id}: {requirement['Name']}\n")
                if requirement['Description']:
                    description = f"**{requirement['RequirementTypeName']}:** {requirement['Description']}\n"
                    formatted_specification.append(description)    
                formatted_specification.append('\n')

                # See if we have any scenarios for this requirement
                _add_requirement_scenarios(spira_client, product_id, requirement_id, formatted_specification)

                # See if we have any defined test cases for this requirement
                # We disabled this and instead now return the test cases in a separate call
                # as it was taking too long and exceeding the MCP Server timeout in Kiro
                # _add_requirement_test_cases(spira_client, product_id, requirement_id, formatted_specification)

        return "\n".join(formatted_specification)
    
    except Exception as e:
        return f"There was a problem using this tool: {e}"
    
def _get_specification_design_impl(spira_client, product_id: int, release_id: int | None) -> str:
    """
    Implementation of retrieving the design markdown specification for the specified product

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
        release_id: The numeric ID of the release. If the ID is RL:12, just use 12.
                    If no release is specified, then the specification for the entire
                    project is returned 
                
    Returns:
        Formatted string containing the product design specification
    """
    try:
        formatted_specification = []

        # Get the product information
        logger.info("Getting the product overview")
        product = _get_product_by_id(spira_client, product_id)
        product_name = product['Name']

        # Create the header
        if release_id:
            # Get the release information
            release = _get_release_by_id(spira_client, product_id, release_id)
            release_version_number = release['VersionNumber']
            formatted_specification.append(f'# Specification for {product_name} [PR:{product_id}], Release {release_version_number} [RL:{release_id}]')
        else:
            formatted_specification.append(f'# Specification for {product_name} [PR:{product_id}]')
        formatted_specification.append('\n')

        # Populate the product overview
        if product['Description']:
            formatted_specification.append(f'## Product Overview')
            formatted_specification.append(product['Description'])
            formatted_specification.append('\n')

        # Create the sub-header for the Design.md section
        formatted_specification.append('\n')
        formatted_specification.append(f'## Design Document')

        # Get the list of risks in the product, or just the release
        risks = _get_specification_risks(spira_client, product_id, release_id)

        if risks:
            # Format the risks into human readable data
            for risk in risks:
                risk_id = risk['RiskId']
                formatted_specification.append(f"### Risk RK:{risk_id}: {risk['Name']}\n")
                if risk['Description']:
                    description = f"**{risk['RiskTypeName']}:** {risk['Description']}\n"
                    formatted_specification.append(description)    
                formatted_specification.append('\n')

                # See if we have any mitigations for this risk
                _add_risk_mitigations(spira_client, product_id, risk_id, formatted_specification)

        return "\n".join(formatted_specification)
    
    except Exception as e:
        return f"There was a problem using this tool: {e}"
    
def _get_specification_tasks_impl(spira_client, product_id: int, release_id: int | None) -> str:
    """
    Implementation of retrieving the tasks markdown specification for the specified product

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
        release_id: The numeric ID of the release. If the ID is RL:12, just use 12.
                    If no release is specified, then the specification for the entire
                    project is returned 
                
    Returns:
        Formatted string containing the product tasks specification
    """
    try:
        formatted_specification = []

        # Get the product information
        logger.info("Getting the product overview")
        product = _get_product_by_id(spira_client, product_id)
        product_name = product['Name']

        # Create the header
        if release_id:
            # Get the release information
            release = _get_release_by_id(spira_client, product_id, release_id)
            release_version_number = release['VersionNumber']
            formatted_specification.append(f'# Specification for {product_name} [PR:{product_id}], Release {release_version_number} [RL:{release_id}]')
        else:
            formatted_specification.append(f'# Specification for {product_name} [PR:{product_id}]')
        formatted_specification.append('\n')

        # Populate the product overview
        if product['Description']:
            formatted_specification.append(f'## Product Overview')
            formatted_specification.append(product['Description'])
            formatted_specification.append('\n')

        # Get the list of requirements in the product, or just the release
        requirements = _get_specification_requirements(spira_client, product_id, release_id)

        # Create the sub-header for the Tasks.md section
        formatted_specification.append('\n')
        formatted_specification.append(f'## Implementation Plan')

        # Loop through the requirements and add the tasks
        if requirements:
            # Format the requirements into human readable data
            for requirement in requirements:
                requirement_id = requirement['RequirementId']
                formatted_specification.append(f"### Requirement RQ:{requirement_id}: {requirement['Name']}\n")

                # See if we have any defined tasks for this requirement
                _add_requirement_tasks(spira_client, product_id, requirement_id, formatted_specification)

        return "\n".join(formatted_specification)
    
    except Exception as e:
        return f"There was a problem using this tool: {e}"
    
def _get_specification_test_cases_impl(spira_client, product_id: int, release_id: int | None) -> str:
    """
    Implementation of retrieving the test cases markdown specification for the specified product

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
        release_id: The numeric ID of the release. If the ID is RL:12, just use 12.
                    If no release is specified, then the specification for the entire
                    project is returned 
                
    Returns:
        Formatted string containing the product test cases specification
    """
    try:
        formatted_specification = []

        # Get the product information
        logger.info("Getting the product overview")
        product = _get_product_by_id(spira_client, product_id)
        product_name = product['Name']

        # Create the header
        if release_id:
            # Get the release information
            release = _get_release_by_id(spira_client, product_id, release_id)
            release_version_number = release['VersionNumber']
            formatted_specification.append(f'# Specification for {product_name} [PR:{product_id}], Release {release_version_number} [RL:{release_id}]')
        else:
            formatted_specification.append(f'# Specification for {product_name} [PR:{product_id}]')
        formatted_specification.append('\n')

        # Populate the product overview
        if product['Description']:
            formatted_specification.append(f'## Product Overview')
            formatted_specification.append(product['Description'])
            formatted_specification.append('\n')

        # Create the sub-header for the test-cases.md section
        formatted_specification.append('\n')
        formatted_specification.append(f'## Test Cases')

        # Get the list of test cases in the product, or just the release
        test_cases = []
        starting_row = 1
        number_of_rows = 250
        more_results = True
        sort_field = 'TestCaseId'
        sort_direction = 'ASC'        
        while more_results:
            test_cases_url = f"projects/{product_id}/test-cases?starting_row={starting_row}&number_of_rows={number_of_rows}&sort_field={sort_field}&sort_direction={sort_direction}&release_id={release_id}"
            results = spira_client.make_spira_api_get_request(test_cases_url)
            if not results:
                more_results = False
            else:
                starting_row += number_of_rows
            test_cases.extend(results)

        if test_cases:
            # Format the test cases into human readable data
            for test_case_item in test_cases:
                # Get the full details of the test case (with steps)
                test_case_id = test_case_item['TestCaseId']
                test_case_url = f"projects/{product_id}/test-cases/{test_case_id}"
                test_case = spira_client.make_spira_api_get_request(test_case_url)
                
                if test_case:
                    name = test_case['Name']
                    formatted_specification.append(f"##### Test Case TC:{test_case_id}: {test_case['Name']}\n")
                    if test_case['Description']:
                        description = f"**{test_case['TestCaseTypeName']}:** {test_case['Description']}\n"
                        formatted_specification.append(description)    
                    formatted_specification.append('\n')

                    # Get the test case steps
                    test_steps_url = f"projects/{product_id}/test-cases/{test_case_id}/test-steps"
                    test_steps = spira_client.make_spira_api_get_request(test_steps_url)

                    # Format the test steps as a table
                    if test_steps:
                        formatted_specification.append('##### Steps\n\n')
                        formatted_specification.append("<table>")
                        formatted_specification.append("<tr><th>Step #</th><th>Description</th><th>Expected Result</th><th>Sample Data</th></tr>")
                        for test_step in test_steps:
                            test_step_id = test_step['TestStepId']
                            position = test_step['Position']
                            description = test_step['Description']
                            expected_result = test_step['ExpectedResult']
                            sample_data = test_step['SampleData']
                            formatted_specification.append("<tr>")
                            formatted_specification.append(f"<td>{position}.</td>")
                            formatted_specification.append(f"<td>{description}.</td>")
                            formatted_specification.append(f"<td>{expected_result}.</td>")
                            formatted_specification.append(f"<td>{sample_data}.</td>")
                            formatted_specification.append("</tr>")
                        formatted_specification.append("</table>")

        return "\n".join(formatted_specification)
    
    except Exception as e:
        return f"There was a problem using this tool: {e}"

def register_tools(mcp) -> None:
    """
    Register product specification tools with the MCP server.
    
    Args:
        mcp: The FastMCP server instance
    """

    @mcp.tool()
    def get_specification_requirements(product_id: int, release_id: int | None) -> str:
        """
        Retrieves the requirements specification file for the requested Spira product,
        with the option to only return the specification for a selected product
        release.
        
        Use this tool when you need to download the requirements part of a product specification
        so that it can be used in an agentic development environment such as Amazon Kiro

        Args:
            product_id: The numeric ID of the product. If the ID is PR:45, just use 45.
            release_id: The numeric ID of the release. If the ID is RL:12, just use 12.
                        If no release is specified, then the specification for the entire
                        project is returned 
        
        Returns:
            Formatted string in markdown that contains the requirements specification for the requested
            Spira product (or just the specific release in that product).

            The data returned should be saved into a file called requirements.md
        """
        try:
            spira_client = get_spira_client()
            return _get_specification_requirements_impl(spira_client, product_id, release_id)
        except Exception as e:
            return f"Error: {str(e)}"
        
    @mcp.tool()
    def get_specification_design(product_id: int, release_id: int | None) -> str:
        """
        Retrieves the design specification file for the requested Spira product,
        with the option to only return the specification for a selected product
        release.
        
        Use this tool when you need to download the design part of a product specification
        so that it can be used in an agentic development environment such as Amazon Kiro

        Args:
            product_id: The numeric ID of the product. If the ID is PR:45, just use 45.
            release_id: The numeric ID of the release. If the ID is RL:12, just use 12.
                        If no release is specified, then the specification for the entire
                        project is returned 
        
        Returns:
            Formatted string in markdown that contains the design specification for the requested
            Spira product (or just the specific release in that product).

            The data returned should be saved into a file called design.md
        """
        try:
            spira_client = get_spira_client()
            return _get_specification_design_impl(spira_client, product_id, release_id)
        except Exception as e:
            return f"Error: {str(e)}"
        
    @mcp.tool()
    def get_specification_tasks(product_id: int, release_id: int | None) -> str:
        """
        Retrieves the tasks specification file for the requested Spira product,
        with the option to only return the specification for a selected product
        release.
        
        Use this tool when you need to download the tasks part of a product specification
        so that it can be used in an agentic development environment such as Amazon Kiro

        Args:
            product_id: The numeric ID of the product. If the ID is PR:45, just use 45.
            release_id: The numeric ID of the release. If the ID is RL:12, just use 12.
                        If no release is specified, then the specification for the entire
                        project is returned 
        
        Returns:
            Formatted string in markdown that contains the tasks specification for the requested
            Spira product (or just the specific release in that product).

            The data returned should be saved into a file called tasks.md
        """
        try:
            spira_client = get_spira_client()
            return _get_specification_tasks_impl(spira_client, product_id, release_id)
        except Exception as e:
            return f"Error: {str(e)}"

    @mcp.tool()
    def get_specification_test_cases(product_id: int, release_id: int | None) -> str:
        """
        Retrieves the test cases specification file for the requested Spira product,
        with the option to only return the specification for a selected product
        release.
        
        Use this tool when you need to download the test cases part of a product specification
        so that it can be used in an agentic development environment such as Amazon Kiro
        This file will include:
            - Detailed test scenarios and steps
            - Expected results and pass/fail criteria
            - Test data requirements
            - Edge cases and negative testing scenarios
            - Performance test specifications

        Args:
            product_id: The numeric ID of the product. If the ID is PR:45, just use 45.
            release_id: The numeric ID of the release. If the ID is RL:12, just use 12.
                        If no release is specified, then the specification for the entire
                        project is returned 
        
        Returns:
            Formatted string in markdown that contains the test cases specification for the requested
            Spira product (or just the specific release in that product).

            The data returned should be saved into a file called test-cases.md
        """
        try:
            spira_client = get_spira_client()
            return _get_specification_test_cases_impl(spira_client, product_id, release_id)
        except Exception as e:
            return f"Error: {str(e)}"


"""
Tests for the Inflectra Spira MCP Server.
"""
import pytest
from mcp.shared.memory import (
    create_connected_server_and_client_session as client_session,
)

from mcp_server_spira.server import mcp

# Mark all tests with anyio for async testing
@pytest.mark.anyio
async def test_server_initialization():
    """Test that the server initializes correctly and returns capabilities."""
    async with client_session(mcp._mcp_server) as client:
        # Initialize the connection
        init_result = await client.initialize()
        
        # Check that initialization was successful
        assert init_result is not None
        
        # Check server name in serverInfo
        assert init_result.serverInfo.name == "Inflectra Spira"
        
        # Check that the server has capabilities
        capabilities = init_result.capabilities
        assert capabilities is not None
        
        # Check for specific capabilities we expect
        assert capabilities.prompts is not None
        assert capabilities.resources is not None
        assert capabilities.tools is not None


"""
Provides operations for recording the results of CI/CD builds into Spira

This module provides MCP tools for recording the results of continuous integration / continuous deployment
pipeline builds against a matching release in Spira 
"""

import datetime

from mcp_server_spira.features.common import get_spira_client

def _create_build_url_impl(spira_client, product_id: int, release_id: int, build_status_id: int, name: str, description: str, commits: list[str]) -> str:
    """
    Creates a new CI/CD pipeline build entry in Spira

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PG:45, just use 45.
        release_id: The ID of the release/sprint/phase in Spira that the build is for, without the RL prefix (e.g. RL:12 would be 12)
        build_status_id: The status of the build (1=Failed, 2=Passed)
        name: The name of the build (usually containing the project name and the date/time of the build)
        description: The detailed description of the build (optional), what was included and why
        commits: An optional array/list of the Git hashes of the commits included in the build
                
    Returns:
        The ID of the new build that was created (with 'BL' prefix)
    """
    try:
        # Populate the revisions object from the commits
        revisions = []
        if commits:
            for commit in commits:
                revision = {
                    'RevisionKey': commit
                }
                revisions.append(revision)

        # The body we are sending
        body = {
            # 1=Failed, 2=Passed
            'ProjectId': product_id,
            'BuildStatusId': build_status_id,
            'ReleaseId': release_id,
            'Name': name,
            'Description': description,
            'Revisions': revisions            
        }

        # Record the test run using the API method
        create_build_url = "projects/" + str(product_id) + "/releases/" + str(release_id) + "/builds "
        build = spira_client.make_spira_api_post_request(create_build_url, body)

        if not build:
            return "The build was not recorded successfully."

        # Extract the new build ID
        build_id = build['BuildId']

        return "BL:" + str(build_id)
    except Exception as e:
        return f"There was a problem using this tool: {e}"
    
def register_tools(mcp) -> None:
    """
    Register tools with the MCP server.
    
    Args:
        mcp: The FastMCP server instance
    """

    @mcp.tool()
    def create_build(product_id: int, release_id: int, build_status_id: int, name: str, description: str, commits: list[str]) -> str:
        """
        Creates a new CI/CD pipeline build entry in Spira
        
        Use this tool when you need to:
        - Push the results of an automated software build into Spira
                    
        Args:
            product_id: The numeric ID of the product. If the ID is PG:45, just use 45.
            release_id: The ID of the release/sprint/phase in Spira that the build is for, without the RL prefix (e.g. RL:12 would be 12)
            build_status_id: The status of the build (1=Failed, 2=Passed)
            name: The name of the build (usually containing the project name and the date/time of the build)
            description: The detailed description of the build (optional), what was included and why
            commits: An optional array/list of the Git hashes of the commits included in the build
                
        Returns:
            The ID of the new build that was created (with 'BL' prefix)
        """
        try:
            spira_client = get_spira_client()
            return _create_build_url_impl(spira_client, product_id, release_id, build_status_id, name, description, commits)
        except Exception as e:
            return f"Error: {str(e)}"
        


"""
Provides operations for working with the Spira program workspace

This module provides MCP tools for retrieving and updating programs (also known as projects).
"""

from mcp_server_spira.features.formatting import format_program
from mcp_server_spira.features.common import get_spira_client

def _get_programs_impl(spira_client) -> str:
    """
    Implementation of retrieving the list of Spira programs
    the current user has access to

    Args:
        spira_client: The Inflectra Spira API client instance
                
    Returns:
        Formatted string containing the list of available programs
    """
    try:
        # Get the list of available programs for the current user
        programs_url = "programs"
        programs = spira_client.make_spira_api_get_request(programs_url)

        if not programs:
            return "There are no programs available for the current user."

        # Format the programs into human readable data
        formatted_results = []
        for program in programs[:100]:  # Only show first 100 programs
            program_info = format_program(program)
            formatted_results.append(program_info)

        return "\n\n".join(formatted_results)
    except Exception as e:
        return f"There was a problem using this tool: {e}"
    
def register_tools(mcp) -> None:
    """
    Register my work tools with the MCP server.
    
    Args:
        mcp: The FastMCP server instance
    """

    @mcp.tool()
    def get_programs() -> str:
        """
        Retrieves a list of the programs (projects) that the current user has access to
        
        Use this tool when you need to:
        - View the list of programs that a user has access to
        - Get information about multiple programs at once
        - Access the full description and selected fields of programs
                    
        Returns:
            Formatted string containing comprehensive information for the
            requested list of programs, including name, id, description and key fields,
            formatted as markdown with clear section headings
        """
        try:
            spira_client = get_spira_client()
            return _get_programs_impl(spira_client)
        except Exception as e:
            return f"Error: {str(e)}"
        


"""
Provides operations for working with the Spira requirements I have been assigned

This module provides MCP tools for retrieving and updating my assigned requirements.
"""

from mcp_server_spira.features.formatting import format_requirement
from mcp_server_spira.features.common import get_spira_client

def _get_my_requirements_impl(spira_client) -> str:
    """
    Implementation of retrieving my assigned Spira requirements.

    Args:
        spira_client: The Inflectra Spira API client instance
                
    Returns:
        Formatted string containing the list of assigned requirements
    """
    try:
        # Get the list of open requirements for the current user
        requirements_url = "requirements"
        requirements = spira_client.make_spira_api_get_request(requirements_url)

        if not requirements:
            return "The current user does not have any requirements."

        # Format the requirements into human readable data
        formatted_results = []
        for requirement in requirements[:25]:  # Only show first 25 requirements
            requirement_info = format_requirement(requirement)
            formatted_results.append(requirement_info)

        return "\n\n".join(formatted_results)
    except Exception as e:
        return f"There was a problem using this tool: {e}"
    
def register_tools(mcp) -> None:
    """
    Register my work tools with the MCP server.
    
    Args:
        mcp: The FastMCP server instance
    """

    @mcp.tool()
    def get_my_requirements() -> str:
        """
        Retrieves a list of the open requirements that are assigned to me
        
        Use this tool when you need to:
        - View the complete details of a specific requirement
        - Examine the current state, assigned user, and other properties
        - Get information about multiple requirements at once
        - Access the full description and selected fields of requirements
                    
        Returns:
            Formatted string containing comprehensive information for the
            requested list of requirements, including name, id, description and key fields,
            formatted as markdown with clear section headings
        """
        try:
            spira_client = get_spira_client()
            return _get_my_requirements_impl(spira_client)
        except Exception as e:
            return f"Error: {str(e)}"
        


"""
Inflectra Spira client utilities.

This module provides helper functions for connecting to Inflectra Spira.
"""
import os
import httpx
from typing import Optional, Tuple, Any

# Constants
USER_AGENT = "mcp-server/1.0"
API_ENDPOINT_URL = "/Services/v7_0/RestService.svc"

def get_base_url() -> Optional[str]:
    """
    Gets the Inflectra Spira base URL from environment variables
    
    Returns:
        String containing the base URL for your instance of Inflectra Spira
    """    
    base_url = os.environ.get("INFLECTRA_SPIRA_BASE_URL")
    return base_url

def get_credentials() -> Tuple[Optional[str], Optional[str]]:
    """
    Get Inflectra Spira credentials from environment variables.
    
    Returns:
        Tuple containing (username, api_key)
    """
    username = os.environ.get("INFLECTRA_SPIRA_USERNAME")
    api_key = os.environ.get("INFLECTRA_SPIRA_API_KEY")
    return username, api_key
        
class SpiraClient:
    def __init__(self, base_url, username, api_key):
        self.base_url = base_url
        self.username = username
        self.api_key = api_key

    def make_spira_api_get_request(self, url: str) -> dict[str, Any] | list[Any] | None:
        """
        Makes an HTTP GET request to the Spira REST API with proper error handling.

        Args:
            url: The Relative URL for the specific REST resouce being called
                    
        Returns:
            List or Dictionary containing the JSON returned from the API
        """

        headers = {
            "User-Agent": USER_AGENT,
            "Accept": "application/json",
            "Content-Type": "application/json",
            "username": self.username,
            "api-key": self.api_key
        }

        # Check that we have the appropriate settings populated
        if self.base_url is None:
            raise ValueError("INFLECTRA_SPIRA_BASE_URL needs to be populated as an environment variable!")
        if self.username is None:
            raise ValueError("INFLECTRA_SPIRA_USERNAME needs to be populated as an environment variable!")
        if self.api_key is None:
            raise ValueError("INFLECTRA_SPIRA_API_KEY needs to be populated as an environment variable!")        

        full_url = self.base_url + API_ENDPOINT_URL + '/' + url
        
        with httpx.Client() as client:
            try:
                response = client.get(full_url, headers=headers, timeout=30.0)
                response.raise_for_status()
                return response.json()
            except Exception as e:
                raise Exception(f"Error returned when calling the Spira REST API. The error message was: {e}")

    def make_spira_api_post_request(self, url: str, json: dict[str, Any] | list[Any]) -> dict[str, Any] | list[Any] | None:
        """
        Makes an HTTP POST request to the Spira REST API with proper error handling.

        Args:
            url: The Relative URL for the specific REST resouce being called
            json: The JSON body of the POST request being sent to the REST resource
                    
        Returns:
            List or Dictionary containing the JSON returned from the API
        """

        headers = {
            "User-Agent": USER_AGENT,
            "Accept": "application/json",
            "Content-Type": "application/json",
            "username": self.username,
            "api-key": self.api_key
        }

        # Check that we have the appropriate settings populated
        if self.base_url is None:
            raise ValueError("INFLECTRA_SPIRA_BASE_URL needs to be populated as an environment variable!")
        if self.username is None:
            raise ValueError("INFLECTRA_SPIRA_USERNAME needs to be populated as an environment variable!")
        if self.api_key is None:
            raise ValueError("INFLECTRA_SPIRA_API_KEY needs to be populated as an environment variable!")        

        full_url = self.base_url + API_ENDPOINT_URL + '/' + url
        
        with httpx.Client() as client:
            try:
                response = client.post(url=full_url, json=json, headers=headers, timeout=30.0)
                response.raise_for_status()
                return response.json()
            except Exception as e:
                raise Exception(f"Error returned when calling the Spira REST API. The error message was: {e}")

    def make_spira_api_put_request(self, url: str, json: dict[str, Any] | list[Any]) -> dict[str, Any] | list[Any] | None:
        """
        Makes an HTTP PUT request to the Spira REST API with proper error handling.

        Args:
            url: The Relative URL for the specific REST resouce being called
            json: The JSON body of the POST request being sent to the REST resource
                    
        Returns:
            List or Dictionary containing the JSON returned from the API
        """

        headers = {
            "User-Agent": USER_AGENT,
            "Accept": "application/json",
            "Content-Type": "application/json",
            "username": self.username,
            "api-key": self.api_key
        }

        # Check that we have the appropriate settings populated
        if self.base_url is None:
            raise ValueError("INFLECTRA_SPIRA_BASE_URL needs to be populated as an environment variable!")
        if self.username is None:
            raise ValueError("INFLECTRA_SPIRA_USERNAME needs to be populated as an environment variable!")
        if self.api_key is None:
            raise ValueError("INFLECTRA_SPIRA_API_KEY needs to be populated as an environment variable!")        

        full_url = self.base_url + API_ENDPOINT_URL + '/' + url
        
        with httpx.Client() as client:
            try:
                response = client.put(url=full_url, json=json, headers=headers, timeout=30.0)
                response.raise_for_status()
                return response.json()
            except Exception as e:
                raise Exception(f"Error returned when calling the Spira REST API. The error message was: {e}")
            
    def make_spira_api_delete_request(self, url: str) -> dict[str, Any] | list[Any] | None:
        """
        Makes an HTTP DELETE request to the Spira REST API with proper error handling.

        Args:
            url: The Relative URL for the specific REST resouce being called
                    
        Returns:
            List or Dictionary containing the JSON returned from the API
        """

        headers = {
            "User-Agent": USER_AGENT,
            "Accept": "application/json",
            "Content-Type": "application/json",
            "username": self.username,
            "api-key": self.api_key
        }

        # Check that we have the appropriate settings populated
        if self.base_url is None:
            raise ValueError("INFLECTRA_SPIRA_BASE_URL needs to be populated as an environment variable!")
        if self.username is None:
            raise ValueError("INFLECTRA_SPIRA_USERNAME needs to be populated as an environment variable!")
        if self.api_key is None:
            raise ValueError("INFLECTRA_SPIRA_API_KEY needs to be populated as an environment variable!")        

        full_url = self.base_url + API_ENDPOINT_URL + '/' + url
        
        with httpx.Client() as client:
            try:
                response = client.delete(url=full_url, headers=headers, timeout=30.0)
                response.raise_for_status()
                return response.json()
            except Exception as e:
                raise Exception(f"Error returned when calling the Spira REST API. The error message was: {e}")

def get_client() -> SpiraClient:

    # Get the base url, login and api key
    base_url = get_base_url()
    username, api_key = get_credentials()

    # Create the Spira client
    spira_client = SpiraClient(base_url, username, api_key)
    return spira_client



"""
Specifications tools for Spira by Inflectra
"""
from mcp_server_spira.features.specifications.tools import (
    productspecification
)


def register_tools(mcp) -> None:
    """
    Register all specification tools with the MCP server.
    
    Args:
        mcp: The FastMCP server instance
    """
    productspecification.register_tools(mcp)
    


"""
Provides operations for getting a list of artifact types and sub-types in the current product template

This module provides MCP tools for retrieving artifact types, and their assoicated sub types
"""

from mcp_server_spira.features.formatting import format_milestone
from mcp_server_spira.features.common import get_spira_client

def _get_artifact_types_impl(spira_client, template_id: int) -> str:
    """
    Implementation of retrieving the list of artifact types and sub-types in the product template

    Args:
        spira_client: The Inflectra Spira API client instance
        template_id: The numeric ID of the product template. If the ID is PT:45, just use 45. 
                
    Returns:
        Formatted string containing the list of artifact types and sub-types
    """
    try:
        formatted_results = "# Artifact Types\n\n"


        # --- Requirements ---
        formatted_results += "## Requirement"
            
        types_url = "project-templates/" + str(template_id) + "/requirements/types"
        types = spira_client.make_spira_api_get_request(types_url)

        if not types:
            return "Unable to fetch requirement types for the product template."

        # Format the types into human readable data
        type_results = []
        for type in types:
            milestone_info = f"""   - {type['Name']} (ID={type['RequirementTypeId']})"""
            type_results.append(milestone_info)

        formatted_results += "\n".join(type_results)    
        formatted_results += "\n\n------------------------------\n\n"


        # --- Test Cases ---
        formatted_results += "## Test Case"
            
        types_url = "project-templates/" + str(template_id) + "/test-cases/types"
        types = spira_client.make_spira_api_get_request(types_url)

        if not types:
            return "Unable to fetch test case types for the product template."

        # Format the types into human readable data
        type_results = []
        for type in types:
            milestone_info = f"""   - {type['Name']} (ID={type['TestCaseTypeId']})"""
            type_results.append(milestone_info)

        formatted_results += "\n".join(type_results)    
        formatted_results += "\n\n------------------------------\n\n"

        # --- Tasks ---
        formatted_results += "## Task"
            
        types_url = "project-templates/" + str(template_id) + "/tasks/types"
        types = spira_client.make_spira_api_get_request(types_url)

        if not types:
            return "Unable to fetch task types for the product template."

        # Format the types into human readable data
        type_results = []
        for type in types:
            milestone_info = f"""   - {type['Name']} (ID={type['TaskTypeId']})"""
            type_results.append(milestone_info)

        formatted_results += "\n".join(type_results)
        formatted_results += "\n\n------------------------------\n\n"

        # --- Risks ---
        formatted_results += "## Risk"
            
        types_url = "project-templates/" + str(template_id) + "/risks/types"
        types = spira_client.make_spira_api_get_request(types_url)

        if not types:
            return "Unable to fetch risk types for the product template."

        # Format the types into human readable data
        type_results = []
        for type in types:
            milestone_info = f"""   - {type['Name']} (ID={type['RiskTypeId']})"""
            type_results.append(milestone_info)

        formatted_results += "\n".join(type_results)
        formatted_results += "\n\n------------------------------\n\n"


        # --- Incidents ---
        formatted_results += "## Incident"
            
        types_url = "project-templates/" + str(template_id) + "/incidents/types"
        types = spira_client.make_spira_api_get_request(types_url)

        if not types:
            return "Unable to fetch incident types for the product template."

        # Format the types into human readable data
        type_results = []
        for type in types:
            milestone_info = f"""   - {type['Name']} (ID={type['IncidentTypeId']})"""
            type_results.append(milestone_info)

        formatted_results += "\n".join(type_results)
        formatted_results += "\n\n------------------------------\n\n"


        # --- Documents ---
        formatted_results += "## Document"
            
        types_url = "project-templates/" + str(template_id) + "/document-types?active_only=true"
        types = spira_client.make_spira_api_get_request(types_url)

        if not types:
            return "Unable to fetch document types for the product template."

        # Format the types into human readable data
        type_results = []
        for type in types:
            milestone_info = f"""   - {type['Name']} (ID={type['DocumentTypeId']})"""
            type_results.append(milestone_info)

        formatted_results += "\n".join(type_results)
        formatted_results += "\n\n------------------------------\n\n"

        return formatted_results
    
    except Exception as e:
        return f"There was a problem using this tool: {e}"

def register_tools(mcp) -> None:
    """
    Register artifact type tools with the MCP server.
    
    Args:
        mcp: The FastMCP server instance
    """

    @mcp.tool()
    def get_artifact_types(template_id: int) -> str:
        """
        Retrieves a list of the artifact types and associated sub-types for the current product template
        
        Use this tool when you need to:
        - View the list of artifact types in the product template
        - For each artifact type (e.g. test case), get the list of sub-types (e.g. test case types)
        - Access the name and ID of each type

        Args:
            template_id: The numeric ID of the product template. If the ID is PT:45, just use 45. 
        
        Returns:
            Formatted string containing comprehensive information for the
            requested list of artifact types and corresponding sub-types
            formatted as markdown with clear section headings
        """
        try:
            spira_client = get_spira_client()
            return _get_artifact_types_impl(spira_client, template_id)
        except Exception as e:
            return f"Error: {str(e)}"
        


"""
Provides operations for working with the Spira product workspace

This module provides MCP tools for retrieving and updating products (also known as projects).
"""

from mcp_server_spira.features.formatting import format_product
from mcp_server_spira.features.common import get_spira_client

def _get_product_by_id_impl(spira_client, product_id: int) -> str:
    """
    Implementation of retrieving a single Spira product by its ID

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45.
                
    Returns:
        Formatted string containing the product definition
    """
    try:
        # Get the product by its ID
        product_url = f"projects/{product_id}"
        product = spira_client.make_spira_api_get_request(product_url)

        if not product:
            return "There was no product with that ID available"

        # Format the product into human readable data
        product_info = format_product(product)

        return product_info
    except Exception as e:
        return f"There was a problem using this tool: {e}"


def _get_products_impl(spira_client) -> str:
    """
    Implementation of retrieving the list of Spira products (projects)
    the current user has access to

    Args:
        spira_client: The Inflectra Spira API client instance
                
    Returns:
        Formatted string containing the list of available products
    """
    try:
        # Get the list of available products for the current user
        products_url = "projects"
        products = spira_client.make_spira_api_get_request(products_url)

        if not products:
            return "There are no products available for the current user."

        # Format the products into human readable data
        formatted_results = []
        for product in products[:100]:  # Only show first 100 products
            product_info = format_product(product)
            formatted_results.append(product_info)

        return "\n\n".join(formatted_results)
    except Exception as e:
        return f"There was a problem using this tool: {e}"

def _get_program_products_impl(spira_client, program_id: int) -> str:
    """
    Implementation of retrieving the list of Spira products (projects)
    the current user has access to

    Args:
        spira_client: The Inflectra Spira API client instance
        program_id: The numeric ID of the program. If the ID is PG:45, just use 45.
                
    Returns:
        Formatted string containing the list of available products
    """
    try:
        # Get the list of available products for the current user
        products_url = "projects"
        products = spira_client.make_spira_api_get_request(products_url)

        if not products:
            return "The program does not contain any products."

        # Loop through and only include the products that are part of the specified program
        # Format the products into human readable data
        formatted_results = []
        for product in products:
            if product['ProjectGroupId'] == program_id:
                product_info = format_product(product)
                formatted_results.append(product_info)

        return "\n\n".join(formatted_results)
    except Exception as e:
        return f"There was a problem using this tool: {e}"
    
def register_tools(mcp) -> None:
    """
    Register my work tools with the MCP server.
    
    Args:
        mcp: The FastMCP server instance
    """

    @mcp.tool()
    def get_products() -> str:
        """
        Retrieves a list of the products (projects) that the current user has access to
        
        Use this tool when you need to:
        - View the list of products that a user has access to
        - Get information about multiple products at once
        - Access the full description and selected fields of products
                    
        Returns:
            Formatted string containing comprehensive information for the
            requested list of products, including name, id, description and key fields,
            formatted as markdown with clear section headings
        """
        try:
            spira_client = get_spira_client()
            return _get_products_impl(spira_client)
        except Exception as e:
            return f"Error: {str(e)}"
    
    @mcp.tool()
    def get_product_by_id(product_id: int) -> str:
        """
        Retrieves a single product by its ID value
        
        Use this tool when you need to:
        - View the details of a single product
        - Access the full description and selected fields of products

        Args:
            product_id: The numeric ID of the product. If the ID is PR:45, just use 45.
                            
        Returns:
            Formatted string containing comprehensive information for the
            requested product, including name, id, description and key fields,
            formatted as markdown with clear section headings
        """
        try:
            spira_client = get_spira_client()
            return _get_product_by_id_impl(spira_client, product_id)
        except Exception as e:
            return f"Error: {str(e)}"

    @mcp.tool()
    def get_program_products(program_id: int) -> str:
        """
        Retrieves a list of the products (projects) that belong to the specified program
        
        Use this tool when you need to:
        - View the list of products that belong to a specific program
        - Get information about multiple products at once
        - Access the full description and selected fields of products

        Args:
            program_id: The numeric ID of the program. If the ID is PG:45, just use 45.

        Returns:
            Formatted string containing comprehensive information for the
            requested list of products, including name, id, description and key fields,
            formatted as markdown with clear section headings
        """
        try:
            spira_client = get_spira_client()
            return _get_program_products_impl(spira_client, program_id)
        except Exception as e:
            return f"Error: {str(e)}"


# Automation feature package for Inflectra Spira MCP
from mcp_server_spira.features.automation import tools


def register(mcp):
    """
    Register all automation components with the MCP server.
    
    Args:
        mcp: The FastMCP server instance
    """
    tools.register_tools(mcp)


"""
Provides operations for getting a list of custom properties defined in the current product template

This module provides MCP tools for retrieving artifact types, and their associated custom properties
"""

from mcp_server_spira.features.formatting import format_milestone
from mcp_server_spira.features.common import get_spira_client

def _get_custom_properties_impl(spira_client, template_id: int) -> str:
    """
    Implementation of retrieving the list of artifact types and custom properties in the product template

    Args:
        spira_client: The Inflectra Spira API client instance
        template_id: The numeric ID of the product template. If the ID is PT:45, just use 45. 
                
    Returns:
        Formatted string containing the list of artifact types and associated custom properties
    """

    formatted_results = "# Artifact Types\n\n"

    # --- Requirements ---
    formatted_results += "## Requirement"
    custom_prop_results = _get_custom_properties_for_artifact_type(spira_client, template_id, 'Requirement')
    formatted_results += custom_prop_results

    # --- Releases ---
    formatted_results += "## Release"
    custom_prop_results = _get_custom_properties_for_artifact_type(spira_client, template_id, 'Release')
    formatted_results += custom_prop_results

    # --- Test Cases ---
    formatted_results += "## Test Case"
    custom_prop_results = _get_custom_properties_for_artifact_type(spira_client, template_id, 'TestCase')
    formatted_results += custom_prop_results

    # --- Tasks ---
    formatted_results += "## Task"
    custom_prop_results = _get_custom_properties_for_artifact_type(spira_client, template_id, 'Task')
    formatted_results += custom_prop_results

    # --- Risks ---
    formatted_results += "## Risk"
    custom_prop_results = _get_custom_properties_for_artifact_type(spira_client, template_id, 'Risk')
    formatted_results += custom_prop_results

    # --- Incidents ---
    formatted_results += "## Incident"
    custom_prop_results = _get_custom_properties_for_artifact_type(spira_client, template_id, 'Incident')
    formatted_results += custom_prop_results

    # --- Test Sets ---
    formatted_results += "## Test Set"
    custom_prop_results = _get_custom_properties_for_artifact_type(spira_client, template_id, 'TestSet')
    formatted_results += custom_prop_results

    # --- Test Steps ---
    formatted_results += "## Test Step"
    custom_prop_results = _get_custom_properties_for_artifact_type(spira_client, template_id, 'TestStep')
    formatted_results += custom_prop_results

    # --- Test Runs ---
    formatted_results += "## Test Run"
    custom_prop_results = _get_custom_properties_for_artifact_type(spira_client, template_id, 'TestRun')
    formatted_results += custom_prop_results

    # --- Automation Hosts ---
    formatted_results += "## Automation Host"
    custom_prop_results = _get_custom_properties_for_artifact_type(spira_client, template_id, 'AutomationHost')
    formatted_results += custom_prop_results

    # --- Documents ---
    formatted_results += "## Documents"
    custom_prop_results = _get_custom_properties_for_artifact_type(spira_client, template_id, 'Document')
    formatted_results += custom_prop_results

    return formatted_results

def _get_custom_properties_for_artifact_type(spira_client, template_id: int, artifact_type_name: str) -> str:

    try:
        custom_props_url = "project-templates/" + str(template_id) + "/custom-properties/" + artifact_type_name
        custom_props = spira_client.make_spira_api_get_request(custom_props_url)

        if not custom_props:
            return ""

        # Format the custom prop into human readable data
        custom_prop_results = []
        for custom_prop in custom_props:
            custom_prop_info = f"""   {custom_prop['PropertyNumber']}. {custom_prop['Name']} (ID={custom_prop['CustomPropertyId']})"""
            custom_prop_results.append(custom_prop_info)

        formatted_results = "\n".join(custom_prop_results)    
        formatted_results += "\n\n------------------------------\n\n"

        return formatted_results
    
    except Exception as e:
        return ""

def register_tools(mcp) -> None:
    """
    Register custom property tools with the MCP server.
    
    Args:
        mcp: The FastMCP server instance
    """

    @mcp.tool()
    def get_custom_properties(template_id: int) -> str:
        """
        Retrieves a list of the artifact types and associated custom properties for the current product template
        
        Use this tool when you need to:
        - View the list of artifact types in the product template
        - For each artifact type (e.g. test case), get the list of custom properties
        - Access the name and ID of each type

        Args:
            template_id: The numeric ID of the product template. If the ID is PT:45, just use 45. 
        
        Returns:
            Formatted string containing comprehensive information for the
            requested list of artifact types and corresponding custom properties
            formatted as markdown with clear section headings
        """
        try:
            spira_client = get_spira_client()
            return _get_custom_properties_impl(spira_client, template_id)
        except Exception as e:
            return f"Error: {str(e)}"
        


"""
Provides operations for working with the Spira product test sets

This module provides MCP tools for retrieving and updating product test sets
"""

from mcp_server_spira.features.formatting import format_test_set
from mcp_server_spira.features.formatting import format_test_set_folder
from mcp_server_spira.features.common import get_spira_client

def _get_test_sets_impl(spira_client, product_id: int) -> str:
    """
    Implementation of retrieving the list of test sets in the specified product, grouped by folder

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
                
    Returns:
        Formatted string containing the list of test sets, grouped by folder
    """
    try:
        formatted_results = []

        # First we need to get any test sets not in a folder
        _get_test_sets_in_folder(spira_client, product_id, formatted_results, None)

        # Get the entire test set folder hierarchy in the product
        test_set_folders_url = f"projects/{product_id}/test-set-folders"
        test_set_folders = spira_client.make_spira_api_get_request(test_set_folders_url)

        # Loop through all the test set folders and format
        for test_set_folder in test_set_folders:
            test_set_folder_info = format_test_set_folder(test_set_folder)
            formatted_results.append(test_set_folder_info)

            # Now get all of the test sets in the folder
            _get_test_sets_in_folder(spira_client, product_id, formatted_results, test_set_folder)

        return "\n\n".join(formatted_results)
    
    except Exception as e:
        return f"There was a problem using this tool: {e}"

def _get_test_sets_in_folder(spira_client, product_id: int, formatted_results: list[str], test_set_folder):
    """
    Implementation of retrieving the list of test sets in the specified product, grouped by folder

    Args:
        spira_client: The Inflectra Spira API client instance
        product_id: The numeric ID of the product. If the ID is PR:45, just use 45. 
                
    Returns:
        Formatted string containing the list of test sets, grouped by folder
    """
    try:
        # Get the test set folder id
        test_set_folder_id = 'null'
        release_id = 'null'
        if test_set_folder:
            test_set_folder_id = str(test_set_folder['TestSetFolderId'])

        # Get the test sets in the folder
        test_sets_url = f"projects/{product_id}/test-set-folders/{test_set_folder_id}/test-sets?starting_row=1&number_of_rows=1000&sort_field=Name&sort_direction=ASC&release_id={release_id}"
        test_sets = spira_client.make_spira_api_get_request(test_sets_url)

        # Loop through all the test sets and format
        for test_set in test_sets:
            test_set_info = format_test_set(test_set)
            formatted_results.append(test_set_info)

    except Exception as e:
        raise Exception(f"Error returned when getting the test sets in the folder. The error message was: {e}")

def register_tools(mcp) -> None:
    """
    Register product test sets tools with the MCP server.
    
    Args:
        mcp: The FastMCP server instance
    """

    @mcp.tool()
    def get_test_sets(product_id: int) -> str:
        """
        Retrieves a list of the test sets in the specified product, grouped by folder
        
        Use this tool when you need to:
        - View the list of test sets in the specified product
        - Get information about multiple test sets at once
        - Access the full description and selected fields of test sets

        Args:
            product_id: The numeric ID of the product. If the ID is PG:45, just use 45. 
        
        Returns:
            Formatted string containing comprehensive information for the
            requested list of test sets, including name, id, description and key fields,
            grouped by the test set folder, formatted as markdown with clear section headings
        """
        try:
            spira_client = get_spira_client()
            return _get_test_sets_impl(spira_client, product_id)
        except Exception as e:
            return f"Error: {str(e)}"
